// (C) 2001-2021 Intel Corporation. All rights reserved.
// Your use of Intel Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files from any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Intel Program License Subscription 
// Agreement, Intel FPGA IP License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Intel and sold by 
// Intel or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


`timescale 1ns/1ps

module intel_pcie_dm_controller_sc_fifo #(
  
  parameter FIFO_DATA_WIDTH = 32,
  parameter FIFO_DEPTH = 8

)
(
    input logic                                       clk,
    input logic                                       rst_n,
    
    input logic [FIFO_DATA_WIDTH-1:0]                 wdata_i,
    input logic                                       write_i,
    input logic                                       read_i,
    
    output logic [FIFO_DATA_WIDTH-1:0]                desc_rdata_o,
    output logic                                      full_o,
    output logic                                      empty_o,
    output logic [FIFO_DEPTH:0]                       fillcount_o


);


logic [FIFO_DATA_WIDTH-1:0]          fifo_reg[(1<<FIFO_DEPTH)-1:0];
logic [FIFO_DEPTH-1:0]               wr_ptr;
logic [FIFO_DEPTH-1:0]               rd_ptr;

logic                             fill_less_than4;


  assign full_o = fillcount_o[FIFO_DEPTH];
  
  always @ (posedge clk) begin
    if (write_i)
      fifo_reg[wr_ptr] <= wdata_i ;
  end 

  assign desc_rdata_o = fifo_reg[rd_ptr];

  always_ff @(posedge clk) begin
    if (!rst_n) begin
      wr_ptr <= '0;
      rd_ptr <= '0;
    end
    else begin
      if (write_i) begin
        wr_ptr <= wr_ptr + 1'b1;
      end
      if (read_i) begin
        rd_ptr <= rd_ptr +  1'b1;
      end
    end
  end
  
  always_ff @(posedge clk) begin
    if (!rst_n) begin
      fillcount_o <= '0;
    end
    else begin
      if (write_i & !read_i)
        fillcount_o <= fillcount_o + 1'b1;
      if (read_i & !write_i)
        fillcount_o <= fillcount_o - 1'b1;
    end
  end

  always_ff @(posedge clk) begin
    if (!rst_n) begin
      fill_less_than4          <= 1'b1;
      empty_o                    <= 1'b1;
    end
    else begin
      fill_less_than4 <= (fillcount_o < 3'd4);
      empty_o <= fill_less_than4 & (((fillcount_o[1:0] == 1) & read_i & !write_i) | (fillcount_o[1:0] == 0));
    end
  end
 
endmodule